function [lone,late, lonr,latr, pm,pn, ang,orterr] = compute_grid( nx,ny, ...
                   size_x,size_y, cent_lat, NStpr,EWtpr, cushn1,xcr1,ycr1, ...
                     cushn2,xcr2,ycr2, cushn3,xcr3,ycr3, cushn4,xcr4,ycr4, ...
                         tra_lon,tra_lat,alpha, flip_xy, save_mode )

% Produce a logically rectangular curvilinear orthogonal grid with minimal
% distortion. Computation involves 5 stages:

% (1) Set up a patch of uniform-resolution Cartesian grid on Mercator plane
%     centered at the point corresponding to the projection of user-specified
%     point with lat=cent_lat on the Greenwich meridian, and with boundaries
%     corresponding to the user-specified "size_y" [km] in north-south and
%     "size_x" (also km] in east-west direction measured along the longitudinal
%     line passing through the center of the patch.   Specifying cent_lat=0
%     makes the patch centered around Equator; a positive(negative) value moves
%     it to the north(south). When projected back to the sphere this makes the
%     grid tapered toward one of the poles with to dx,dy~cos(lat-cent_lat),
%     where dx,dx are measured in meters.  An infinitely small patch is almost
%     rectangular if placed on Equator, while behaving more and more like
%     a sector of polar coordinates if placed closer to one of the poles.

% (2) Optionally apply a set of successive conformal transformations to change
%     the rectangle on Mercator plane into a desired shape.

% (3) Transfer from Mercator plane to the sphere [x,y] --> [lon,lat];

% (4) move the grid toward Equator as a solid object, azimuthally rotate it
%     around it center by angle "alpha", then moved to the north or south to
%     latitude "tra_lat", and finally moved east or west to longitude
%     "tra_lon".

% (5) Compute metric terms, angle between local positive X-direction and
%     direction to the east, and orthogonality error. 

%   inputs: nx,ny: numbers of grid points in XI- and ETA-directions (here X-Y)
%                  counting only inner RHO-points (boundary rows are excluded)
%                  same as LLm and MMm in ROMS code, with the exception of the
%                  possibility to switch their roles if "flip_xy" is set to
%                  transpose the grid;

%   size_x,size_y: domain size in X- and Y-direction measured in [km];

%        cent_lat: latitude of of the initial location of the grid center;
%     NStpr,EWtpr: parameter for north-south and east-west tapering;

% cushn1,xcr1,ycr1: parameters for cushioning conformal transformations:
% cushn2,xcr2,ycr2: strength and coordinates of the center (no stretching or
% cushn3,xcr3,ycr3: squeezing takes place at the center, while it increases
% cushn4,xcr4,ycr4: linearly with the distance from the center in a small
%                   vicinity of it;

%         tra_lat: desired latitude of grid center;
%         tra_lon: desired longitude of grid center;
%           alpha: azimuthal angle of grid direction (alpha=0 means that
%                  positive X-direction at the center of the grid points
%                  exactly to the east);

%         flip_xy: flag to transpose the grid: 0 do nothing; 1 flip X-direction
%                  and transpose; 2 flip Y-direction and transpose.  Thus 1 and
%                  2 corresponds to 90-degree azimuthal rotation in counter- or
%                  or clockwise directions (respectively) however this rotation
%                  is "logical" rather than physical, i.e., the two are the
%                  same for would be perfectly rectangular grid, but are not
%                  equivalent in the case of spherical.

%      save_mode:  switch to select between plotting only (save_mode == 0) and  
%                  generating the actual 2D grid with the intent to save it
%                  into grid file (save_mode == 1).  This is merely to speed up
%                  the responsiveness of GUI update because for the plotting
%                  purposes all what is needed is perimeter of lonr,latr, but
%                  not lone,late, pm,pn, ang,orterr.


 disp(['compute_grid :: nx=',num2str(nx),          ' ny=',num2str(ny), ...
                  ' size_x=',num2str(size_x),  ' size_y=',num2str(size_y) ])

 disp(['                cent_lat=',num2str(cent_lat), ...
                    ' NStpr=',num2str(NStpr),    ' EWtpr=',num2str(EWtpr) ])

 disp(['                cushn1=',num2str(cushn1), ' xcr1=',num2str(xcr1), ...
                                                  ' ycr1=',num2str(ycr1), ...
                      ' cushn2=',num2str(cushn2), ' xcr2=',num2str(xcr2), ...
                                                  ' ycr2=',num2str(ycr2) ])

 disp(['                cushn3=',num2str(cushn3), ' xcr3=',num2str(xcr3), ...
                                                  ' ycr3=',num2str(ycr3), ...
                      ' cushn4=',num2str(cushn4), ' xcr4=',num2str(xcr4), ...
                                                  ' ycr4=',num2str(ycr4) ])

 disp(['                tra_lon=',num2str(tra_lon), ...
                   ' tra_lat=',num2str(tra_lat), ' alpha=',num2str(alpha), ...
          ' flip_xy=',num2str(flip_xy), ' save_mode=',num2str(save_mode) ])


%% Stage 1: Create a patch of Cartesian grid on Mercator plane (x,y).
%% ======== At first, find the extents xmin,xmax,ymin,ymax corresponding
%%          to user-specified sizes and center latitude, then create the
%%          patch, then extract its perimeter (if for plotting only). 

  pi=3.14159265358979323846; deg2rad=pi/180; Eradius = 6371315.;

  theta=deg2rad*cent_lat ;        %% latitude of grid center in radians
  dtht=0.5*size_y/Eradius ;       %% half of north-south extent in radians
  ymax=lat_to_eta(theta+dtht);    %% north and south bounds of the initial
  ymin=lat_to_eta(theta-dtht);    %% rectangular domain on Mercator plane

  ycent=0.5*(ymin+ymax);          %% center on Mercator plane
  ysize=ymax-ymin;
  dy=ysize/double(ny);            %% Temporarily set north and south bounds
  ymin=-0.5*ysize;                %% symmetrically about the Equator - their
  ymax=+0.5*ysize;                %% original values can be reinstated by
                                  %% adding ycent to both.

  x_size=size_x/(Eradius*cos(theta)) ;     %% east-west extents of
  xmin=-0.5*x_size ; xmax=+0.5*x_size ;    %% the grid expressed in
  dx=x_size/double(nx) ;                   %% radians

  disp(['ISOTROPY ERROR (dx/dy)-1 = ', num2str(dx/dy -1.) ])

  xr=zeros(nx+2,ny+2); yr=zeros(nx+2,ny+2);
  for j=1:ny+2                                 %% Create a patch of
    for i=1:nx+2                               %% uniform rectangular
      xr(i,j)=xmin+dx*(double(i)-1.5);         %% Cartesian grid on
      yr(i,j)=ymin+dy*(double(j)-1.5);         %% Mercator plane,
    end                                        %% RHO-points.
  end
  if (save_mode == 1)
    xe=zeros(nx+3,ny+3); ye=zeros(nx+3,ny+3);  %% same as above, but
    for j=1:ny+3                               %% for PSI-points
      for i=1:nx+3
        xe(i,j)=xmin+dx*double(i-2);
        ye(i,j)=ymin+dy*double(j-2);
      end                                      %% Because for the purpose
    end                                        %% of plotting all what is 
  else                                         %% needed is double-line 
    [xe,ye, xr,yr] = perim2(xr,yr);            %% perimeter, extract it 
  end                                          %% and discard the interior.


%% Stage 2: Apply a set of successive conformal transforms to achieve
%% ======== the desired shape of grid perimeter (still on Mercator plane)

%% Select spatial scale common to both horizontal directions

%%  xy_size=sqrt((xmax-xmin)*(ymax-ymin));

                                          %% Harmonic mean: no more than
  xy_size=2.*(xmax-xmin)*(ymax-ymin) ...  %% twice the smaller of either
           /((xmax-xmin)+(ymax-ymin)) ;   %% (xmax-xmin) or (ymax-ymin).

  xc1=(xmin+xmax)/2. + xcr1*(xmax-xmin)/2.;
  yc1=(ymin+ymax)/2. + ycr1*(ymax-ymin)/2.;
  xc2=(xmin+xmax)/2. + xcr2*(xmax-xmin)/2.;
  yc2=(ymin+ymax)/2. + ycr2*(ymax-ymin)/2.;

  xc3=(xmin+xmax)/2. + xcr3*(xmax-xmin)/2.;
  yc3=(ymin+ymax)/2. + ycr3*(ymax-ymin)/2.;
  xc4=(xmin+xmax)/2. + xcr4*(xmax-xmin)/2.;
  yc4=(ymin+ymax)/2. + ycr4*(ymax-ymin)/2.;

  [xe,ye] = cushion(xe,ye, cushn1, xc1,yc1, xy_size);
  [xe,ye] = cushion(xe,ye, cushn2, xc2,yc2, xy_size);
  [xe,ye] = cushion(xe,ye, cushn3, xc3,yc3, xy_size);
  [xe,ye] = cushion(xe,ye, cushn4, xc4,yc4, xy_size);

  [xe,ye] = polarX(xe,ye, EWtpr, xmin,xmax,ymin,ymax);
  [xe,ye] = polarY(xe,ye, NStpr, xmin,xmax,ymin,ymax);
      ye  =  ye+ycent;

  [xr,yr] = cushion(xr,yr, cushn1, xc1,yc1, xy_size);
  [xr,yr] = cushion(xr,yr, cushn2, xc2,yc2, xy_size);
  [xr,yr] = cushion(xr,yr, cushn3, xc3,yc3, xy_size);
  [xr,yr] = cushion(xr,yr, cushn4, xc4,yc4, xy_size);

  [xr,yr] = polarX(xr,yr, EWtpr, xmin,xmax,ymin,ymax);
  [xr,yr] = polarY(xr,yr, NStpr, xmin,xmax,ymin,ymax);
      yr  =  yr+ycent;

%% Stage 3: transfer from Mercator plane to the sphere [x,y] --> [lon,lat];
%% ======= 
                            %% at extended PSI-points
  late=asin(tanh(ye));      %% latitude in radians
  lone=xe;                  %% longitude in radians

  latr=asin(tanh(yr));      %% at RHO-points
  lonr=xr;                  %% keep in radians
                            %% longitude in radians
  clear xe ye xr yr ;

  if (flip_xy > 0)
    lone = flipdim(lone,flip_xy)'; lonr = flipdim(lonr,flip_xy)';
    late = flipdim(late,flip_xy)'; latr = flipdim(latr,flip_xy)';
  end



%% Stage 4: move the grid to the desired place and rotate it azimuthally 
%% ========

  [lonr,latr] = move_and_turn(tra_lon,tra_lat, alpha, cent_lat, lonr,latr);
  [lone,late] = move_and_turn(tra_lon,tra_lat, alpha, cent_lat, lone,late);


%% Stage 5: Compute angles of local grid positive x-axis relative to east,
%% ======== metric coefficients pm,pn, check for othogonality;  

 [ii,jj]=size(latr) ; pm=zeros(ii,jj) ; ang=zeros(ii,jj) ;
                      pn=zeros(ii,jj) ; orterr=zeros(ii,jj) ;

 if (save_mode == 1)
   disp('Calculating angle ...')


%% ang=0.5*cos(latr);  %% <-- temporally
%%
%% a11 = ang .* ( lone(2:ii+1,2:jj+1) -lone(1:ii,2:jj+1) ... % cos(Lat)*dLon/dX
%%               +lone(2:ii+1,1:jj  ) -lone(1:ii,1:jj  ) ) ;
%%
%% a12 = ang .* ( lone(1:ii,2:jj+1) +lone(2:ii+1,2:jj+1) ... % cos(Lat)*dLon/dY
%%               -lone(1:ii,1:jj  ) -lone(2:ii+1,1:jj  ) ) ;
%%
%% a21 =  0.5 * ( late(2:ii+1,2:jj+1) -late(1:ii,2:jj+1) ... %     dLat/dX
%%               +late(2:ii+1,1:jj  ) -late(1:ii,1:jj  ) ) ;
%%
%% a22 =  0.5 * ( late(1:ii,2:jj+1) +late(2:ii+1,2:jj+1) ... %     dLat/dY
%%               -late(1:ii,1:jj  ) -late(2:ii+1,1:jj  ) ) ;


   for j=1:jj
     for i=1:ii
       dLnX1=lone(i+1,j+1)-lone(i,j+1);        % Most of the time "lone"
       if (dLnX1 > pi)                         % is expected to be a smooth
         dLnX1=dLnX1-2*pi;                     % function of its indices,
       elseif (dLnX1 < -pi)                    % however it may experience
         dLnX1=dLnX1+2*pi;                     % 2*pi jumps due to periodicity
       end                                     % resulting in large vales of
       dLnX=lone(i+1,j)-lone(i,j);             % differences dLnX,dLnY. If this
       if (dLnX > pi)                          % happens, 2*pi is added or
         dLnX=dLnX-2*pi;                       % subtracted as appropriate.
       elseif (dLnX < -pi)
         dLnX=dLnX+2*pi;
       end

       dLnY1=lone(i+1,j+1)-lone(i+1,j);        % Because of this possibility,
       if (dLnY1 > pi)                         % the elegant Matlab-style array
         dLnY1=dLnY1-2*pi;                     % syntax code above is commented
       elseif (dLnY1 < -pi)                    % out permanently, but still
         dLnY1=dLnY1+2*pi;                     % left here for Matlab lovers
       end                                     % to admire, while quasi-Fortran
       dLnY=lone(i,j+1)-lone(i,j);             % code of the left does the job.
       if (dLnY > pi)
         dLnY=dLnY-2*pi;
       elseif (dLnY < -pi)
         dLnY=dLnY+2*pi;
       end

       cff=0.5*cos(latr(i,j));
       a11=cff*(dLnX+dLnX1);                    % cos(Lat)*dLon/dX

       a12=cff*(dLnY+dLnY1);                    % cos(Lat)*dLon/dY

       a21=0.5*( late(i+1,j+1) -late(i,j+1) ... %     dLat/dX
                +late(i+1,j  ) -late(i,j  ) ) ;

       a22=0.5*( late(i,j+1) +late(i+1,j+1) ... %     dLat/dY
                -late(i,j  ) -late(i+1,j  ) ) ;
                                               % Note that this computation
       pm(i,j)=1./(Eradius*sqrt(a11^2+a21^2)); % of pm and pn overwrites their
       pn(i,j)=1./(Eradius*sqrt(a12^2+a22^2)); % previously computed analytical
                                               % counterparts.
       if (a21 < -abs(a11))
         ang1 = -0.5*pi-atan(a11/a21);         % For a perfectly othogonal
       elseif (a21 > abs(a11))                 % curvilinear grid "ang1" and
         ang1 =  0.5*pi-atan(a11/a21);         % "ang2" should be the same.
       else                                    % Here we chose to compute both
         ang1 = atan(a21/a11);                 % of them, use their average as
         if (a11 < 0.)                         % the final accepted value for
           if (a21 < 0.)                       % east angle and also compute
             ang1=ang1-pi;                     % and save their difference as
           else                                % the measure for orthogonality
             ang1=ang1+pi;                     % error.
           end
         end
       end

       if (a12 < -abs(a22))
         ang2 = 0.5*pi +atan(a22/a12);
       elseif (a12 > abs(a22))
         ang2 = -0.5*pi +atan(a22/a12);
       else
         ang2 = -atan(a12/a22);
         if (a22 < 0.)
           if (a12 < 0.)
             ang2 = ang2 +pi;
           else
             ang2 = ang2 -pi;
           end
         end
       end

       ang(i,j)=0.5*(ang1+ang2);  %% the two should be same or very close
       orterr(i,j) = ang1-ang2 ;  %% orthogonality error

     end  %% <-- for i=1:ii
   end  %% <-- for j=1:jj
   disp([ 'compute_grid :: orthogonality error=', ...
      num2str(180*max(max(abs(orterr)))/pi), ' degrees' ])
   disp('fully updated compute_grid')
 else
   disp('updated lon,lat in compute_grid')
 end  %% <-- if save_mode
end %%<-- function compute_grid


function eta = lat_to_eta(theta)
  pi=3.14159265358979323846;
  if (-0.5*pi < theta && theta < 0.5*pi)
    cff=sin(theta);
    eta=0.5*log( (1.+cff)/(1.-cff) );
  else
    disp(['lat_to_eta :: theta=', num2str(theta)])
    error('### ERROR: Latitude range exception.')
  end
end

function [x1,y1, x2,y2] = perim2(x,y)
  disp('perim2 :: extracting perimeter')
  [nx,ny]=size(x);
  x1=[ x(1:nx-1,1)' x(nx,1:ny-1)   x(nx:-1:2,ny)'     x(1,ny:-1:2) ];
  y1=[ y(1:nx-1,1)' y(nx,1:ny-1)   y(nx:-1:2,ny)'     y(1,ny:-1:2) ];

  x2=[ x(2:nx-2,2)' x(nx-1,2:ny-2) x(nx-1:-1:3,ny-1)' x(2,ny-1:-1:3) ];
  y2=[ y(2:nx-2,2)' y(nx-1,2:ny-2) y(nx-1:-1:3,ny-1)' y(2,ny-1:-1:3) ];
end

function [x1,y1] = cushion(x,y, cushn, x0,y0,xy_size)
  pi=3.14159265358979323846;
  if (cushn > 0.)
    cff=cushn*pi/xy_size; cff1=1./cff;
    argX=cff*x0; argY=cff*y0;
    xoffst=cff1*sin(argX)*cosh(argY);
    yoffst=cff1*cos(argX)*sinh(argY);

    argX=cff*(x-x0); argY=cff*(y-y0);
    x1=xoffst + cff1 * sin(argX) .* cosh(argY);
    y1=yoffst + cff1 * cos(argX) .* sinh(argY);

  elseif (cushn < 0.)
    cff=-cushn*pi/xy_size; cff1=1./cff;
    argX=cff*x0; argY=cff*y0;
    xoffst=cff1*sinh(argX)*cos(argY);
    yoffst=cff1*cosh(argX)*sin(argY);

    argX=cff*(x-x0); argY=cff*(y-y0);
    x1=xoffst + cff1*sinh(argX) .* cos(argY);
    y1=yoffst + cff1*cosh(argX) .* sin(argY);

  else
    x1=x;  y1=y;
  end
end

function [x1,y1] = polarX(x,y, EWtpr, xmin,xmax,ymin,ymax)
  if (abs(EWtpr) > 0.0001)
    x0=(xmin+xmax)/2.0 -(xmax-xmin)/EWtpr ;
    rmin=xmin-x0 ; rmax=xmax-x0 ;
    dRdX=( log(rmax/rmin) )/(xmax-xmin ) ;
    y0=0.5*(ymin+ymax) ;

    r=rmin*exp( dRdX*(x-xmin) ) ;
    alpha=dRdX*( y-y0 ) ;
    x1=x0 + r .* cos(alpha) ;
    y1=y0 + r .* sin(alpha) ;
  else
    x1=x;  y1=y;
  end
end

function [x1,y1] = polarY(x,y, NStpr, xmin,xmax,ymin,ymax)
  if (abs(NStpr) > 0.0001)
    y0=(ymin+ymax)/2.D0 -(ymax-ymin)/NStpr ;
    rmin=ymin-y0 ; rmax=ymax-y0 ;
    dRdY=( log(rmax/rmin) )/(ymax-ymin ) ;
    x0=0.5*(xmin+xmax) ;

    r=rmin*exp( dRdY*(y-ymin) ) ;
    alpha=dRdY*(x-x0) ;
    x1=x0 + r .* sin(alpha) ;
    y1=y0 + r .* cos(alpha) ;
  else
    x1=x;  y1=y;
  end
end
